/*  Copyright (C) 2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 4/24/13.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __mc_engine_h__
#define __mc_engine_h__ 1



/*   std includes   */


/*   project headers   */
#include <bits/mc-engine-opts.h>
#include <bits/mc-shower-history.h>
#include <bits/mc-splitting-prob.h>
#include <bits/mc-color-evolution.h>
#include <bits/mc-threshold.h>
#include <bits/qcd-color-overlap.h>
#include <cxx/cloned-ptr.h>



namespace duct {
  
  
  /*  This is the core of the shower evolution code. It uses a given shower probability class to make
   *  trial emission. An instance of the shower probability class is stored in this object.
   */
  template<ProcessTag _Proc, OrderingTag _Ordering>
  class shower_engine
  {
  public:
    /*    constructors  */
    shower_engine() = default;
    
    explicit shower_engine(std::size_t s)
    : _M_random{s} {}
    
    
    /*   Evolution operator of the threshold part.    */
    void threshold_stage(double kTcutoff, double virtcutoff, const shower_engine_opts::threshold_opts& opts, shower_history_stage<_Proc>& stage)
    {
      if(opts.enabled)
      {
        /*   make sure that the hard parameters are set for the hard state   */
        stage.template update_params<_Ordering>(evolution_type::typeI, kTcutoff, virtcutoff);
        
        /*   applying the LC+ threshold operator   */
        averaged_lcplus_threshold_sudakov<_Proc,_Ordering>(stage);
        
        
        /*   transforms color state : psi => xi   */
        stage.next_time = 9.9999e+127;
        
        _M_vcop.template threshold<_Ordering>(opts, _M_random, stage);
        
        /*   the threshold operator doesn't changes the shower time    */
        stage.next_time = stage.current_time;
      }
    }

    
    /*   make a trial emission and transforms the color state    */
    bool shower_stage(double kTcutoff, double virtcutoff, const shower_engine_opts::color_opts& opts, double tmF, shower_history_stage<_Proc>& current)
    {
      /*   no emission is possible    */
      if(tmF <= current.next_time || std::abs(static_cast<double>(current.weights)) < 1e-25 || current.cstate.empty())
        return false;
      
      
      /*   reducing the size of the color state    */
      current.cstate.mc_reduce(1.0e-8, _M_random, 0.8, opts.max_number_of_basis_states);

      if(current.cstate.empty()) {
        current.weights.born.main = current.weights.nlo.main = 0.0;
        return false;
      }

      
      /*   make sure that the hard parameters are set for the current state   */
      current.template update_params<_Ordering>(opts.type_of_evolution, kTcutoff, virtcutoff);
      
      
      /*   make a trial emission   */
      bool is_possible = _M_prob.make_trial(_M_random, current, tmF, current.next_time);
      
      
      /*   transforms color state : psi => xi   */
      auto csize = current.cstate.size();
      
      _M_vcop.template shower<_Ordering>(opts, _M_random, current);
      
      
      /*   do the real emission   */
      if(is_possible)
      {
        /*   try to reduce the color state by MC summation   */
        if(current.cstate.size() > csize)
          current.cstate.mc_reduce(1.0e-8, _M_random, 0.8, opts.max_number_of_basis_states);
        
        if(current.cstate.empty()) {
          current.weights.born.main = current.weights.nlo.main = 0.0;
          return false;
        }

        _M_rcop.template shower<_Ordering>(opts, _M_random, current);
      }
      
      /*   transform to U(3) if it is needed    */
      current.cstate.transform([=, max_color = opts.max_color_suppression](color_state<_Proc>& ci) {
        this->_M_overlap.convert_to_u3(max_color, ci);
      });
      
      /*    cleaning up the color state   */
      current.cstate.reduce(1.0e-8);

      
      /*   null color state stops the shower   */
      if(current.cstate.empty()) {
        current.weights.born.main = current.weights.nlo.main = 0.0;
        return false;
      }
      
      
      return is_possible;
    }
    
    
    /*   Post-shower stage. Here we consider further MC reduction to speed up the user routine.    */
    void postshower_stage(const shower_engine_opts& opts, shower_history_stage<_Proc>& current)
    {
      if(current.cstate.size() > 20u*opts.max_number_of_color_basis_states)
      {
        current.cstate.mc_reduce(1.0e-8, _M_random, 0.8, 2u*opts.max_number_of_color_basis_states);
        
        if(current.cstate.empty())
          current.weights.born.main = current.weights.nlo.main = 0.0;
      }
    }
    
  private:
    /*   random engine   */
    random_engine_type _M_random;
    
 
    /*   the underlying splitting probability   */
    splitting_prob<_Proc, _Ordering> _M_prob;
    
    
    /*    implementation of the real and virtual color operators   */
    real_shower_color_operator<_Proc> _M_rcop;
    virtual_shower_color_operator<_Proc> _M_vcop;
    
    
    /*   calculates color overlaps in the evolution    */
    color_overlap _M_overlap;
  };

  
  
  
  /*  This is the shower evolution code. The type of the ordering class is erased.    */
  template<ProcessTag _Proc>
  class shower_evolution
  {
  public:
    /*    default constructor    */
    shower_evolution() = default;
    
    /*    create the shower engine from a given shower probability   */
    template<OrderingTag _Ordering>
    explicit shower_evolution(_Ordering, const shower_engine_opts& opts, std::size_t s)
    : _M_opts{opts}, _M_trial{new obj<_Ordering>{s}} {}
    
    
    /*   shower evolution    */
    void operator()(shower_history<_Proc>& hist) {
      _M_trial->evolve(_M_opts, hist);
    }
    
    
    /*   Getting engine options     */
    const shower_engine_opts& options() const { return _M_opts;}
    
    
  private:
    /*   engine parameters    */
    shower_engine_opts _M_opts;
    
 
    /*   abstract class for type erasing   */
    struct base {
      virtual void evolve(const shower_engine_opts&, shower_history<_Proc>&) = 0;
    };
    
    /*   this stores an instanse of the actual splitting probability   */
    template<OrderingTag _Ordering>
    struct obj : base
    {
      explicit obj(std::size_t s) : engine{s} {}
      
      void evolve(const shower_engine_opts& opts, shower_history<_Proc>& hist)
      {
        using _Engine = shower_engine<_Proc, _Ordering>;
        
        /*   if it is the first step we have to apply the thrshold operator   */
        hist.threshold_stage(&_Engine::threshold_stage, engine, opts.kT_cutoff, opts.virtuality_cutoff, opts.threshold);
        
        
        /*   shower evolution    */
        for(const auto& [scale, copts] : opts.color)
          while(hist.shower_stage(sqr(scale), &_Engine::shower_stage, engine, opts.kT_cutoff, opts.virtuality_cutoff, copts));
        
        
        /*   post shower stage   */
        // hist.postshower_stage(&_Engine::postshower_stage, engine, opts);
      }
      
      
      /*   the underlying splitting probability   */
      shower_engine<_Proc, _Ordering> engine;
    };
    
    
    /*   we put an instance of obj into a cloned_ptr   */
    cloned_ptr<base> _M_trial;
  };
  
}   /*   namespace duct   */

#endif
