/*  Copyright © 2015 Zoltán Nagy. All rights reserved.
 *  Created by Zoltán Nagy on 28/10/15.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __mc_hard_state_h__
#define __mc_hard_state_h__ 1


/*   std includes   */

/*   project headers   */
#include <bits/mc-shower-history.h>
#include <cxx/cloned-ptr.h>
#include <cxx/random.h>



namespace duct {
  

  /*   Requirements for a hard state   */
  template<typename _Hard, ProcessTag _Proc>
  concept bool HardState = std::is_copy_constructible_v<_Hard>
  && requires(_Hard& h, random_engine_type& rne, shower_stage<_Proc>& stage, shower_vetos& vetos)
  {
    /*   generates a hard event   */
    {h.generate(rne, stage, vetos)} -> void;
  };
  
  
  
  /*   This object is for store (and use) generator of the hard states.
   *   The actual type of this generator is not visible it is erases,
   *   but one istance of the hard state generator is stored in this class.
   */
  template<class _Proc>
  class hard_state
  {
  public:
    /*   default constructor (no managed object)   */
    hard_state() = default;
    hard_state(hard_state&&) = default;
    hard_state(const hard_state&) = default;
    
    /*   create from a hard state generator   */
    template<typename _Hard> requires HardState<_Hard, _Proc>
    hard_state(_Hard&& h, std::size_t s)
    : _M_hard{new obj<_Hard>{std::forward<_Hard>(h), s}} {}
    
    
    /*    generate the hard state    */
    shower_history<_Proc>& operator()(shower_history<_Proc>& hist) {
      _M_hard->generate(hist);
      return hist;
    }
 
    
  private:
    /*   abstract class for type erasing   */
    struct base {
      virtual void generate(shower_history<_Proc>&) = 0;
    };
    
    /*   this stores an instanse of the actual hard process   */
    template<class _Hard>
    struct obj : base
    {
      explicit obj(_Hard&& h, std::size_t s)
      : hard{std::forward<_Hard>(h)}, random{s} {}
      
      /*   generate event for the shower history   */
      void generate(shower_history<_Proc>& hist) {
        hist.hard_stage(&obj::generate_hard_stage, this);
      }

      /*   genrate the hard stage    */
      void generate_hard_stage(shower_history_stage<_Proc>& stage)
      {
        /*   Generating the user visable part of the hard event.
         *   Our requirement to fill the following members of class shower_stage:
         *       ==> state
         *       ==> cstate (only: colorL, colorR, weight)
         *       ==> weight.born.main
         *       ==> weight.nlo.main
         *       ==> scale
         *       ==> vetos
         */
        hard.generate(random, stage, stage.vetos);
        
        /*   setting up the color states  */
        for(auto& c : stage.cstate)
        {
          /*   make the health report   */
          c.make_health_report();
          
          /*   We always start with SU(Nc) states.   */
          c.is_color_u3 = false;
          
          /*   color supression index of the evolution (always zero in the hard state)     */
          c.color_suppression_index = 0u;
          
          /*   color supression index of the overlap     */
          c.overlap_color_suppression_index = color_suppression_index_u3(c);
          
          /*   control variables    */
          c.order_of_soft_corrections = 0u;
          c.order_of_softH_corrections = 0u;
          c.order_of_softV_corrections = 0u;
          c.order_of_softT_corrections = 0u;
          c.order_of_phase_corrections = 0u;
          c.order_of_phaseL_corrections = 0u;
          c.order_of_phaseR_corrections = 0u;
          c.order_of_phaseU_corrections = 0u;
        }
        
        
        /*   reference momentum to define shower time    */
        stage.qnull = momentum_sum<incomings>(stage.state);
        
        /*    shower time    */
        stage.current_time = stage.next_time = std::log(stage.qnull.mag2()/stage.scale);
        
        /*   threshold weights   */
        stage.weights.born.threshold = stage.weights.nlo.threshold = 1.0;
      }

      /*    the underlying hard process    */
     _Hard hard;

      /*   random engine   */
      random_engine_type random;
    };
    
    /*   we put an instance of obj<_Hard> into "any"   */
    cloned_ptr<base> _M_hard;
  };
 
}

#endif
