/*  Copyright (C) 2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 8/24/13.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __mc_user_h__
#define __mc_user_h__ 1


/*   project headers     */
#include <bits/mc-engine.h>
#include <bits/qcd-color-overlap.h>


namespace duct {
  
  /*   Abstract class to define the most general user interface.
   *   This is just the base of the user class, but it is independent 
   *   of the process type.
   */
  class user_base
  {
  public:
    /*   Virtual destructor   */
    virtual ~user_base() = default;
    
    /*   User defined finish function. It is called once
     *   at the end of the calculation.
     */
    virtual void finish() {}
    
    /*   Using this function the deductor kernel can pass some information
     *   such as PID and name of the process. Those are helpful to compose 
     *   output file names. The default does nothing.
     */
    virtual void process_info(unsigned int, const char *) {}

    /*   Make sure that the event counter is set to zero
     *   and the maximum number of events is set correctly.
     */
    void initialize() {
      this->initfunc();
      this->_M_number_of_events = 0;
    }
    
    /*   Increment the event number and do some user defined operations at the
     *   end of the event
     */
    void end_of_event() {
      ++_M_number_of_events;
      this -> operations_at_the_end_of_event();
    }
    
    /*   Returns the number of the analysed events    */
    unsigned long long int number_of_events() const {
      return _M_number_of_events;
    }
    
    /*   returns true if this calculation is finished (no more events are needed)   */
    bool is_done() const {
      return max_number_of_events == 0 ? false : _M_number_of_events >= max_number_of_events;
    }
    
  private:
    /*   User defined init function. It is called once
     *   at the begining of the calculation.
     */
    virtual void initfunc() = 0;
    
    /*   User defined operations at the end of the event. It is virtual but not abstract
     *   function. User might don't have to do anything at the end of the event.
     */
    virtual void operations_at_the_end_of_event() {}
    
  protected:
    /*   maximum number of events to generate  */
    unsigned long long int max_number_of_events = 0;
    
  private:
    /*   number of the analysed events   */
    unsigned long long int _M_number_of_events = 0;
  };
  
  
  /*   Abstract class to define the most general user interface.
   *   Every analyser should be inhereted from this class.
   *   This class adds the userfunc(...) function and defines the
   *   full user<Tag> abstract class.
   */
  template<ProcessTag _Tag>
  class user : public virtual user_base
  {
  public:
    /*   Virtual destructor   */
    virtual ~user() = default;
    
    
    /*   This analyser routine is used in most of the time.
     *   Usualally we need access only to the hard, threshold and showered stages.
     */
    virtual void userfunc(const shower_stage<_Tag>&, const shower_stage<_Tag>&, const shower_stage<_Tag>&) = 0;
  };

 
  
  /*    Helper class for calculate color weights in the user function.
   *    It is used in the basic_user<...> class.
   */
  template<ProcessTag _Tag>
  class user_weights
  {
  public:
    /*   This is the overlap calculator function.    */
    double color_weight(const shower_stage<_Tag>& s) const
    {
      if(s.cstate.empty()) return 0.0;
      
      /*    calculate all the overlaps    */
      double ret = 0.0;
      
      for(auto& ci : s.cstate)
        if(ci.order_of_phase_corrections % 2 == 0)
          ret += (ci.order_of_phase_corrections % 4 ? -ci.weight : ci.weight)*_M_overlap(ci);
      
      return ret;
    }
    
    
    /*   This is the overlap calculator function for a color basis state.    */
    double color_weight(const color_state<_Tag>& ci) const {
      return ci.weight*_M_overlap(ci);
    }
    
    
    /*   returns the full weight   */
    double weight(const shower_stage<_Tag>& s) const {
      return color_weight(s)*s.weights.born.main*s.weights.born.threshold;
    }
    
    
    /*   This is a resonable breakdown of the full weight
     *     0 => Standard weight (no delta, no threshold)
     *     1 => Standard * (Standard Threshold)
     *     2 => (Standard + Delta) (no threshold)
     *     3 => (Standard + Delta + DeltaThreshold)*(Standard Threshold)
     *     4 => (Standard + Delta + Phase) (no threshold)
     *     5 => (Standard + Delta + Phase + DeltaThreshold)*(Standard Threshold)
     */
    std::array<double,6u> weight_info(const shower_stage<_Tag>& s, double fac = 1.0) const
    {
      if(s.cstate.empty()) return {{0.0, 0.0, 0.0, 0.0, 0.0, 0.0}};
      
      /*    calculate all the overlaps    */
      std::array<double,6u> ret;
      double L = 0.0, D = 0.0, T = 0.0, P = 0.0, PT = 0.0;
      
      
      for(auto& ci : s.cstate)
        if(ci.order_of_phase_corrections % 2 == 0)
        {
          double tmp = (ci.order_of_phase_corrections % 4 ? -ci.weight : ci.weight)*_M_overlap(ci);
          
          
          if(ci.order_of_softT_corrections)
          {
            if(ci.order_of_phase_corrections) PT += tmp;
            else T += tmp;
          }
          else {
            if(ci.order_of_phase_corrections)
              P += tmp;
            else if(ci.order_of_soft_corrections)
              D += tmp;
            else L += tmp;
          }
        }
      
      double wborn = s.weights.born.main*fac;
      
      ret[0] = L*wborn;
      ret[1] = ret[0]*s.weights.born.threshold;
      
      ret[2] = ret[0] + D*wborn;
      ret[3] = (ret[2] + T*wborn)*s.weights.born.threshold;
      
      ret[4] = ret[2] + P*wborn;
      ret[5] = (ret[4] + (T+PT)*wborn)*s.weights.born.threshold;
      
      return ret;
    }
    
    
  private:
    /*   every analyser has its own resources to calculate color overlaps   */
    mutable color_overlap _M_overlap;
  };
  
}

#endif
