/*  Copyright © 2016 Zoltán Nagy. All rights reserved.
 *  Created by Zoltán Nagy on 16/02/16.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __qcd_coupling_h__
#define __qcd_coupling_h__ 1


/*   std includes   */
#include <utility>

/*   project headers   */
#include <bits/qcd-color.h>
#include <cxx/binary-io.h>



namespace duct {
  namespace qcd {

    /*    Analytic QCD coupling implementation
     *    We can use this for intialize the grid based coupling
     */
    class coupling
    {
    public:
      /*    Constructor  
       *      => Mz = 91.1876GeV, M2z = 8315.17839376GeV^2
       *      => as(Mz)/(2pi) = 0.1185/(2pi) = 0.01885986075638959729
       *      => mc = 1.3GeV, m2c = 1.69GeV^2
       *      => mb = 4.5GeV, m2b = 20.25GeV^2
       *      => mt = inf
       */
      explicit coupling(qcd::order = qcd::order::nlo,
                        double = 0.01885986075638959729, double = 8315.17839376,
                        double = 1.69, double = 20.25, double = 1e+200);
      
      
      /*   return the coupling and the number of active flavors   */
      std::pair<double, unsigned> asnf(double q2) const
      {
        unsigned int nf = 6;
        while(q2 < _M_thrs[nf] && nf > 3) --nf;
        return {_M_base_as(nf, q2, _M_lmd[nf]), nf};
      }
      
      
      /*    returns just the alpha_s value    */
      double operator()(double q2) const
      {
        unsigned int nf = 6;
        while(q2 < _M_thrs[nf] && nf > 3) --nf;
        return _M_base_as(nf, q2, _M_lmd[nf]);
      }
      
      /*    number of the active flavour as function of the scale    */
      unsigned int number_of_flavors(double q2) const {
        unsigned int nf = 6;
        while(q2 < _M_thrs[nf] && nf > 3) --nf;
        return nf;
      }
      
      /*   threshold information (in GeV^2)   */
      double thrs(unsigned int i) const { return _M_thrs[i];}
      
      /*   get Lambda_QCD  (in GeV^2)   */
      double lmd(unsigned int i) const { return _M_lmd[i];}
      
      /*   order of the evolution   */
      qcd::order order() const { return _M_ord;}
      
      /*   I/O  operations   */
      void read(std::istream& is) {
        unsigned ord;
        duct::read(is, ord, _M_lmd, _M_thrs);
        _M_ord = static_cast<qcd::order>(ord);
      }
      
      void write(std::ostream& os) const {
        duct::write(os, static_cast<unsigned>(_M_ord), _M_lmd, _M_thrs);
      }

    private:
      /*   evolution order   */
      qcd::order _M_ord;
      
      /*   Lambda(nf,MSbar), thresholds   */
      double _M_lmd[7], _M_thrs[7];
      
      /*    Some helper function    */
      double _M_base_as(unsigned int, double, double) const;
      double _M_astolmd(unsigned int, double, double) const;
    };
    
    
    /*   I/O  operations   */
    inline std::istream& read(std::istream& is, coupling& m) {
      m.read(is);
      return is;
    }
    
    inline std::ostream& write(std::ostream& os, const coupling& m) {
      m.write(os);
      return os;
    }

    
  }
}

#endif