/*  Copyright © 2016 Zoltán Nagy. All rights reserved.
 *  Created by Zoltán Nagy on 03/03/16.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __qcd_dglap_h__
#define __qcd_dglap_h__ 1


/*   std includes   */
#include <vector>
#include <functional>
#include <memory>
#include <iostream>


/*   project headers   */
#include <bits/qcd-coupling.h>
#include <cxx/binary-io.h>
#include <cxx/math-support.h>



namespace duct {
  
  namespace __details {
    
    /*    container for the pdf array   */
    class pdf_array
    {
    public:
      /*    constructors and assignment  */
      explicit pdf_array(double x = 0.0) { for(auto& p : _M_elem) p = x;}
      
      pdf_array& operator*=(double x) { for(auto& p : _M_elem) p *= x; return *this;}
      pdf_array& operator/=(double x) { for(auto& p : _M_elem) p /= x; return *this;}
      
      /*    more assignments    */
      pdf_array& operator+=(const pdf_array& x) {
        const double *c = x._M_elem;
        for(int k = 0; k < 9; k++) _M_elem[k] += c[k];
        return *this;
      }
      
      pdf_array& operator-=(const pdf_array& x) {
        const double *c = x._M_elem; for(int k = 0; k < 9; k++) _M_elem[k] -= c[k];
        return *this;
      }
      
      /*   element access   */
      const double& operator[](unsigned int i) const { return _M_elem[i];}
      double& operator[](unsigned int i) { return _M_elem[i];}
      
      /*    Arithmetic operators    */
      pdf_array operator-() const { return pdf_array(*this) *= -1.0;}
      pdf_array operator*(double b) const { return pdf_array(*this) *= b;}
      pdf_array operator/(double b) const { return pdf_array(*this) /= b;}
      
      pdf_array operator+(const pdf_array& b) const { return pdf_array(*this) += b;}
      pdf_array operator-(const pdf_array& b) const { return pdf_array(*this) -= b;}
      
      /*   I/O operations   */
      friend std::istream& read(std::istream& is, pdf_array& a) { return duct::read(is, a._M_elem);}
      friend std::ostream& write(std::ostream& os, const pdf_array& a) { return duct::write(os, a._M_elem);}
      
    private:
      /*   9 elem of the pdf array   */
      double _M_elem[9];
    };
    
    
    /*    container for the splitting pdf_kernels   */
    class pdf_kernel
    {
    public:
      /*   constructors   */
      explicit pdf_kernel(double x = 1.0)  {
        for(auto& p : _M_elem) p = 0.0;
        _M_elem[0] = _M_elem[6] = _M_elem[12] = _M_elem[18] = _M_elem[24] = x;
        _M_elem[25] = _M_elem[26] = x;
      }
      
      /*   assignment   */
      pdf_kernel& operator*=(double x) {
        for(auto& p : _M_elem) p *= x;
        return *this;
      }
      
      pdf_kernel& operator/=(double x) {
        for(auto& p : _M_elem) p /= x;
        return *this;
      }
      
      /*   more assignments   */
      pdf_kernel& operator+=(const pdf_kernel& x) {
        const double *c = x._M_elem;
        for(int k = 0; k < 27; k++) _M_elem[k] += c[k];
        return *this;
      }
      
      pdf_kernel& operator-=(const pdf_kernel& x) {
        const double *c = x._M_elem;
        for(int k = 0; k < 27; k++) _M_elem[k] -= c[k];
        return *this;
      }
      
      /*    element access    */
      const double* operator[](unsigned int i) const { return _M_elem + 5*i;}
      double* operator[](unsigned int i) { return _M_elem+5*i;}
      
      /*    Arithmetic operators    */
      pdf_array operator*(const pdf_array& b) const;
      pdf_kernel operator*(const pdf_kernel& b) const;
      pdf_kernel operator-() const { return pdf_kernel(*this) *= -1.0;}
      pdf_kernel operator*(double b) const { return pdf_kernel(*this) *= b;}
      pdf_kernel operator/(double b) const { return pdf_kernel(*this) /= b;}
      pdf_kernel operator+(const pdf_kernel& b) const { return pdf_kernel(*this) += b;}
      pdf_kernel operator-(const pdf_kernel& b) const { return pdf_kernel(*this) -= b;}
      
      /*    The inverse of the kernel    */
      void inverse();
      
    private:
      /*   5 x 5 elements for the signlet part and 2 elements for the non-singlets   */
      double _M_elem[27];
    };
    
    /*  Some operator what we need to do arithmetics with
     *  object with type pdf_array.
     */
    pdf_kernel inverse(const pdf_kernel&);
    
    inline pdf_array operator*(double b, const pdf_array& a) {
      return pdf_array(a) *= b;
    }
    
    inline pdf_kernel operator*(double b, const pdf_kernel& a) {
      return pdf_kernel(a) *= b;
    }
    
    inline pdf_kernel inverse(const pdf_kernel& x)
    {
      pdf_kernel r{x};
      r.inverse();
      return r;
    }
    
    
    /*   Container class for storing the pdf interpolation table.   */
    class pdf_matrix
    {
    public:
      /*     constructors     */
      pdf_matrix() : _M_nrow{0u}, _M_ncol{0u} {}
      pdf_matrix(std::size_t n, std::size_t m) : _M_nrow{n}, _M_ncol{m}, _M_elem(n*m, pdf_array{}) {}
      
      /*   resizing the matrix    */
      void resize(std::size_t n, std::size_t m) {
        _M_nrow = n; _M_ncol = m;
        _M_elem.resize(n*m, pdf_array{});
      }
      
      /*  Element access: the return of these functions points to
       *  the first element of the kth row. To get an element of the
       *  matrix, one should do just m[r][c].
       */
      pdf_array * operator[](std::size_t k) { return _M_elem.data()+_M_ncol*k;}
      const pdf_array * operator[](std::size_t k) const { return _M_elem.data()+_M_ncol*k;}
      
      /* The size of the matrix */
      std::size_t nrow() const { return _M_nrow;}
      std::size_t ncol() const { return _M_ncol;}
      
      /*   I/O  operations   */
      friend std::istream& read(std::istream& is, pdf_matrix& m) { return duct::read(is, m._M_nrow, m._M_ncol, m._M_elem);}
      friend std::ostream& write(std::ostream& os, const pdf_matrix& m) { return duct::write(os, m._M_nrow, m._M_ncol, m._M_elem);}
      
      /*   swapping two matrices   */
      void swap(pdf_matrix& m) {
        std::swap(_M_nrow, m._M_nrow);
        std::swap(_M_ncol, m._M_ncol);
        _M_elem.swap(m._M_elem);
      }
      
    private:
      /*   data members   */
      std::size_t _M_nrow, _M_ncol;
      std::vector<pdf_array> _M_elem;
    };
    
  }   //   namespace __details
  
  

  namespace qcd {
  

    namespace __qcd_details {
      
      /*    This class can be used to solve the DGLAP evolution equation.
       */
      class dglap_grid
      {
      public:
        /*    constructors   */
        dglap_grid()
        : _M_ord{order::nlo}, _M_m2T{1.0}, _M_q2max{1e+10} {}
        
        dglap_grid(order ord, double m2T, double q2max, coupling&& as)
        : _M_ord{ord}, _M_as{std::move(as)}, _M_m2T{m2T}, _M_q2max{q2max} {}
        
        dglap_grid(order ord, double m2T, double q2max, const coupling& as)
        : _M_ord{ord}, _M_as{as}, _M_m2T{m2T}, _M_q2max{q2max} {}
        
        
        /*   do the evolution  */
        void evolve(const std::function<void(double, double *)>&);
        
        
        /*   calculate the pdf for a given flavor
         *
         *      81, -81 => u+ubar + d+dbar + s+sbar + c+cbar + b+bbar
         *      82, -82 => c+cbar
         *      83, -83 => b+bbar
         */
        double parton(int, double, double) const;
        
        
        /*   I/O  operations   */
        void read(std::istream& is) {
          unsigned ord;
          duct::read(is, ord, _M_as, _M_m2T, _M_q2max, _M_x, _M_t, _M_h);
          _M_ord = static_cast<order>(ord);
        }
        
        void write(std::ostream& os) const {
          duct::write(os, static_cast<unsigned>(_M_ord), _M_as, _M_m2T, _M_q2max, _M_x, _M_t, _M_h);
        }
        
      private:
        /*   order of the PDF evolution    */
        order _M_ord;
        
        /*   strong coupling   */
        coupling _M_as;
        
        /*   transverse momentum cutoff (starting scale)    */
        double _M_m2T, _M_q2max;
        
        /*   t grid     */
        std::vector<double> _M_x;
        std::vector<double> _M_t;
        
        /*   the interpolation tables  */
        __details::pdf_matrix _M_h;
        
        /*   private functions to set up the grid    */
        void _M_xt_grid();
        
        /*    access to the gridpoint    */
        double _M_parton(int, unsigned, unsigned) const;
        
        /*   function to calculate the weights in the grid point    */
        __details::pdf_kernel _M_lo_weights(unsigned, unsigned) const;
        __details::pdf_kernel _M_nlo_weights(unsigned, unsigned, unsigned) const;
        __details::pdf_kernel _M_weights(unsigned, unsigned, unsigned) const;
      };
      
    }   //   namespace __qcd_details
    
    
    
    
    class dglap
    {
    public:
      /*    constructors   */
      dglap() : _M_anti{false}, _M_dglap{nullptr} {}
      
      
      /*     do the evolution initialized by qcd_input    */
      void evolve(int key, const input& __in)
      {
        /*   the qcd running coupling    */
        coupling as{__in.order, __in.asref/(2.0*M_PI), sqr(__in.qref), sqr(__in.mc), sqr(__in.mb), sqr(__in.mt)};
       
        /*   creating the dglap grid object   */
        _M_dglap = std::make_shared<__qcd_details::dglap_grid>(__in.order, sqr(__in.mT), 1.0e+10, as);
        _M_anti = false;
        
        /*   do the evolution   */
        _M_dglap->evolve(__in.pdf.at(key));
      }
      
      /*   do the evolution but the order of the pdf evolution is overriden by the template argument   */
      template<order _Ord>
      void evolve(int key, const input& __in)
      {
        /*   the qcd running coupling    */
        coupling as{__in.order, __in.asref/(2.0*M_PI), sqr(__in.qref), sqr(__in.mc), sqr(__in.mb), sqr(__in.mt)};
        
        /*   creating the dglap grid object   */
        _M_dglap = std::make_shared<__qcd_details::dglap_grid>(_Ord, sqr(__in.mT), 1.0e+10, as);
        _M_anti = false;
        
        /*   do the evolution   */
        _M_dglap->evolve(__in.pdf.at(key));
      }
      
      
      /*   calculate the pdf for a given flavor
       *
       *      81, -81 => u+ubar + d+dbar + s+sbar + c+cbar + b+bbar
       *      82, -82 => c+cbar
       *      83, -83 => b+bbar
       */
      double parton(int f, double x, double q2) const {
        return _M_dglap->parton((_M_anti ? -f : f), x, q2);
      }
      
      /*   turn it to anti hadron   */
      void anti() { _M_anti = !_M_anti;}
      
      
      /*   I/O  operations   */
      void read(std::istream&);
      void write(std::ostream&) const;
      
      
      /*    check if the object is empty   */
      bool is_empty() const { return _M_dglap == nullptr;}
      
    private:
      /*    shared pointer to avoid the same copy of the big matrices in the memory    */
      bool _M_anti;
      std::shared_ptr<__qcd_details::dglap_grid> _M_dglap;
    };
    
    
    /*   I/O  operations   */
    inline std::istream& read(std::istream& is, dglap& m) {
      m.read(is);
      return is;
    }
    
    inline std::ostream& write(std::ostream& os, const dglap& m) {
      m.write(os);
      return os;
    }

    
  }   //   namespace qcd
}   //   namespace duct

#endif


