/*  Copyright (C) 2012-2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 04/29/19.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */


//   std includes
#include <cmath>
#include <algorithm>

//   project headers
#include "proc-hhcjets-real.h"
#include "cxx/math-support.h"



namespace duct {
  namespace __details {
  
    //================= ONLY FOR TESTING ======================================
   
    lorentzvector<double> random_momentum(random_engine_type& rand)
    {
      std::uniform_real_distribution<double> range(0.0, 1.0);
    
      double E = -std::log(range(rand)*range(rand));
      double pz = E*(2*range(rand)-1.);
      double pT = std::sqrt(E*E-pz*pz);
      double phi = 2.0 * M_PI * range(rand);
      
      return {E, pT*std::cos(phi), pT*std::sin(phi), pz};
    }
    
    double rambo_weight(unsigned int n, double s)
    {
      static constexpr double __8pi = 25.13274122871834590768;
      static constexpr double __16pi2 = 157.91367041742973790108;
      
      static const double fact[] = {
        1.000000000000000000000, 1.000000000000000000000,
        1.000000000000000000000, 2.000000000000000000000,
        3.464101615137754587050, 5.241482788417793214280,
        7.325683002969412729490, 9.711867496604076595340,
        12.39635026120588892703, 15.37624008321082805550,
        
        18.64921022211253121271, 22.21334732883035904420,
        26.06704940665039454817, 30.20895473052848875326,
        34.63789090520187258023, 39.35283742689448285515,
        44.35289755618292354658, 49.63727677424801864857,
        55.20526599889455311234, 61.05622831083046240725,
        
        67.189588314900381553350, 73.604823510775700955150,
        80.301457218028968401580, 87.279052719104399855340,
        94.537208367699191038680, 102.07555347055500771508,
        109.89374479486319958781, 117.99146358623666580202,
        126.36841300677002588944, 135.02431592135521673949,
        
        143.95891297472152603605, 153.17196091274811705250,
        162.66323111025862674191, 172.43250827433702043364,
        182.47958929763210591531, 192.80428224046477961708,
        203.40640542405709375678, 214.28578662004717056856,
        225.44226232377745474056, 236.87567710075244487879,
      };
      
      return std::pow(s/(__16pi2*fact[n]), (int) (n-2))/__8pi;
    }
    
   
    double rambo_generator(random_engine_type& rand, event<hhc>& hard)
    {
      auto first = begin<qcd_out>(hard);
      auto last = end<qcd_out>(hard);
      auto it = first;
      
      /* generate momenta */
      lorentzvector<double> temp(0.0, 0.0, 0.0, 0.0);

      while(it != last)
      {
        temp += (it->momentum = __details::random_momentum(rand));
        ++it;
      }
      
      /* set the parameters for conform transformation */
      double shat = (hard[-1].momentum + hard[0].momentum).mag2();
      double x = std::sqrt(shat)/temp.mag();
      threevector bVec = -temp.boostVector();
      
      /* set momenta in cm frame */
      it = first;
      while(it != last)
      {
        (it-> momentum).boost(bVec);
        (it-> momentum).operator*=(x);
        ++it;
      }
      
      /* boosting momenta to lab frame */
      temp = hard[-1].momentum + hard[0].momentum;
      bVec = temp.boostVector();
      it = first;
      
      while(it != last)
      {
        (it-> momentum).boost(bVec);
        ++it;
      }
      
      return __details::rambo_weight(3, shat);
    }
    
    //==========================================================================
    
    
    /*    tr(abcd) = s_ab s_cd + s_ad s_bc - s_ac s_bd   */
    template<std::ptrdiff_t _Low, std::ptrdiff_t _High>
    inline double gamma_trace(const dot_products<_Low, _High>& s, int a, int b, int c , int d) {
      return s[a][b]*s[c][d]+s[a][d]*s[b][c]-s[a][c]*s[b][d];
    };
    
    
    /*    tr(abcdec) = s_ad s_bc s_ce + s_ac s_be s_cd - s_ac s_bd s_ce - s_ae s_bc s_cd   */
    template<std::ptrdiff_t _Low, std::ptrdiff_t _High>
    inline double gamma_trace(const dot_products<_Low, _High>& s, int a, int b, int c, int d, int e) {
      return s[a][d]*s[b][c]*s[c][e] + s[a][c]*s[b][e]*s[c][d] - s[a][c]*s[b][d]*s[c][e] - s[a][e]*s[b][c]*s[c][d];
    }
    
    
    /*    Helper functions for the 5 gluon process
     */
    namespace proc5g {
      
      inline double A1(const dot_products<-1,3>& s, int j1, int j2, int j3, int j4, int j5) {
        return 2.0/(s[j1][j2]*s[j1][j5]*s[j2][j3]*s[j3][j4]*s[j4][j5]);
      }
      
      inline double R2(const dot_products<-1,3>& s, int j1, int j2, int j3, int j4, int) {
        return -0.5*gamma_trace(s, j3, j1, j2, j4)/(s[j1][j3]*s[j2][j4]);
      }
      
      inline double R3(const dot_products<-1,3>& s, int j1, int j2, int j3, int j4, int j5) {
        return 0.5*gamma_trace(s, j3, j1, j2, j4, j5)/(s[j1][j3]*s[j2][j4]*s[j2][j5]);
      }
      
      inline double R4(const dot_products<-1,3>& s, int j1, int j2, int j3, int, int j5) {
        return -0.5*gamma_trace(s, j3, j1, j5, j2)/(s[j1][j3]*s[j2][j5]);
      }
      
      inline double R5(const dot_products<-1,3>& s, int, int j2, int j3, int j4, int j5) {
        return -0.5*gamma_trace(s, j4, j2, j3, j5)/(s[j2][j4]*s[j3][j5]);
      }
      
      inline double R6(const dot_products<-1,3>& s, int j1, int j2, int j3, int j4, int j5) {
        return 0.5*gamma_trace(s, j5, j1, j3, j4, j2)/(s[j1][j3]*s[j2][j4]*s[j3][j5]);
      }
      
      inline double R7(const dot_products<-1,3>& s, int j1, int, int j3, int j4, int j5) {
        return -0.5*gamma_trace(s, j4, j1, j5, j3)/(s[j1][j4]*s[j3][j5]);
      }
      
      inline double R8(const dot_products<-1,3>& s, int j1, int j2, int j3, int j4, int j5) {
        return 0.5*gamma_trace(s, j5, j3, j4, j1, j2)/(s[j1][j4]*s[j2][j4]*s[j3][j5]);
      }
      
      inline double R9(const dot_products<-1,3>& s, int j1, int j2, int j3, int j4, int j5) {
        return 0.5*gamma_trace(s, j1, j4, j5, j2, j3)/(s[j1][j4]*s[j2][j5]*s[j3][j5]);
      }
      
      inline double R10(const dot_products<-1,3>& s, int j1, int j2, int, int j4, int j5) {
        return -0.5*gamma_trace(s, j4, j1, j2, j5)/(s[j1][j4]*s[j2][j5]);
      }
      
      inline double R11(const dot_products<-1,3>& s, int j1, int j2, int j3, int j4, int j5)
      {
        double num = s[j3][j4]*(sqr(s[j1][j2])*s[j3][j5]*s[j4][j5] + sqr(s[j1][j5])*s[j2][j3]*s[j2][j4])
        + s[j1][j5]*(sqr(s[j2][j3])*s[j1][j4]*s[j4][j5] + sqr(s[j3][j4])*s[j1][j2]*s[j2][j5])
        + s[j1][j2]*(s[j1][j3]*(sqr(s[j4][j5])*s[j2][j3] - s[j2][j5]*s[j3][j4]*s[j4][j5])
                     - s[j2][j3]*s[j4][j5]*(s[j3][j4]*s[j1][j5] + s[j1][j4]*s[j3][j5]) - s[j1][j5]*s[j2][j4]*s[j3][j4]*s[j3][j5])
        - s[j1][j5]*s[j2][j3]*(s[j1][j3]*s[j2][j4]*s[j4][j5] + s[j1][j4]*s[j2][j5]*s[j3][j4]);
        
        return 0.5*num/(s[j1][j3]*s[j1][j4]*s[j2][j4]*s[j2][j5]*s[j3][j5]);
      }
      
      inline double R12(const dot_products<-1,3>& s, int j1, int j2, int j3, int j4, int j5) {
        return 0.5*gamma_trace(s, j4, j3, j1, j2, j5)/(s[j1][j3]*s[j1][j4]*s[j2][j5]);
      }
     
      
      
      /*
       *     Five gluon amplitudes
       */
#define S(i, j) kin[p##i][p##j]

      double su3_tree(const dot_products<-1,3>& kin, int p1, int q2, int q3, int q4, int q5)
      {
        static const int perm1[12][4] = {
          {0,1,2,3}, {0,1,3,2}, {0,2,1,3}, {0,2,3,1},
          {0,3,1,2}, {0,3,2,1}, {1,0,2,3}, {1,0,3,2},
          {1,2,0,3}, {1,3,0,2}, {2,0,1,3}, {2,1,0,3}
        };
        

        int p2, p3, p4, p5, p[4] = {q2,q3,q4,q5};
        double sum = 0.0;
        
        for(int i = 0; i < 12; i++)
        {
          p2 = p[perm1[i][0]];
          p3 = p[perm1[i][1]];
          p4 = p[perm1[i][2]];
          p5 = p[perm1[i][3]];
          sum += 1.0/(S(1,2)*S(2,3)*S(3,4)*S(4,5)*S(5,1));
        }
        
        sum *= meta::pow<4>(S(1,2)) + meta::pow<4>(S(1,3))
        +      meta::pow<4>(S(1,4)) + meta::pow<4>(S(1,5))
        +      meta::pow<4>(S(2,3)) + meta::pow<4>(S(2,4))
        +      meta::pow<4>(S(2,5)) + meta::pow<4>(S(3,4))
        +      meta::pow<4>(S(3,5)) + meta::pow<4>(S(4,5));
        
        sum /= qcd::Na2;
        sum /= 8.0;

        
        return 4.0*qcd::Nc2*qcd::Nc*qcd::Na*sum;
      }
    }   //   namespace proc5g
 
    
    
    namespace proc2q3g {
      
      inline double A1(const dot_products<-1,3>& s, int j1, int j2, int j3, int j4, int j5) {
        return 2.0/(s[j1][j2]*s[j1][j5]*s[j2][j3]*s[j3][j4]*s[j4][j5]);
      }
      
      inline double R2(const dot_products<-1,3>& s, int, int j2, int j3, int j4, int j5) {
        return -0.5*gamma_trace(s, j4, j2, j3, j5)/(s[j2][j4]*s[j3][j5]);
      }
      
      inline double R3(const dot_products<-1,3>& s, int j1, int j2, int j3, int j4, int) {
        return -0.5*gamma_trace(s, j3, j1, j2, j4)/(s[j1][j3]*s[j2][j4]);
      }
      
      inline double R4(const dot_products<-1,3>& s, int j1, int j2, int j3, int j4, int j5) {
        return 0.5*gamma_trace(s, j3, j1, j2, j4, j5)/(s[j1][j3]*s[j2][j4]*s[j2][j5]);
      }
      
      inline double R5(const dot_products<-1,3>& s, int j1, int j2, int j3, int j4, int j5) {
        return 0.5*gamma_trace(s, j5, j3, j4, j1, j2)/(s[j1][j4]*s[j2][j4]*s[j3][j5]);
      }
      
      inline double R6(const dot_products<-1,3>& s, int j1, int j2, int, int j4, int j5) {
        return 0.5*gamma_trace(s, j4, j1, j2, j5)/(s[j1][j4]*s[j2][j5]);
      }
  
      
      double su3_tree(const dot_products<-1,3>& kin, int p1, int p2, int p3,int p4,int p5)
      {
        const double s13 = S(1,3), s14 = S(1,4), s15 = S(1,5), s12 = S(1,2),
        s34 = S(3,4), s35 = S(3,5), s32 = S(3,2), s45 = S(4,5),
        s42 = S(4,2), s52 = S(5,2);
        
        double prop = s13*s14*s15*s32*s42*s52;
        double num  = s13*s32*(s13*s13 + s32*s32) + s14*s42*(s14*s14 + s42*s42) + s15*s52*(s15*s15 + s52*s52);
        double sum1 = (s13*s42+s14*s32)/s34 + (s13*s52+s15*s32)/s35 + (s14*s52+s15*s42)/s45;
        
        double sum2 = s13*s32*(s14*s52 + s15*s42)/(s34*s35)
        + s14*s42*(s15*s32 + s13*s52)/(s45*s34)
        + s15*s52*(s13*s42 + s14*s32)/(s45*s35);
        
        return 0.25*qcd::Na*num*(s12 + qcd::Nc2*(s12-sum1) + qcd::Nc2*qcd::Nc2*sum2/s12)/(prop*qcd::Nc2);
      }
    }
    
  }   //  namespace __details
  
  
  
  void hhcjets_real::generate(random_engine_type& rne, shower_stage<hhc>& hard)
  {
    std::uniform_real_distribution<double> rng(0.0, 1.0);
    
    
    /* resizing state */
    hard.state.resize(0, 3);
    hard.cstate.clear();
    
    
    /* incoming hadrons */
    double Ecm = std::sqrt(_M_s);
    
    hard.state[hadron(-1)].momentum = {0.5*Ecm, 0.0, 0.0, 0.5*Ecm};
    hard.state[hadron(0)].momentum = {0.5*Ecm, 0.0, 0.0, -0.5*Ecm};
    hard.state[hadron(-1)].key = _M_hadronA;
    hard.state[hadron(0)].key = _M_hadronB;
    
    

     /*    weight of the event    */
    double weight = 1.0;
    
    
    //================ WE NEED A BETTER PHASE SPACE!!!! ====================
    
    /* parton momenta */
    double eta = rng(rne);
    double etb = rng(rne);
    
    hard.state[-1].momentum = hard.state[hadron(-1)].momentum*eta;;
    hard.state[ 0].momentum = hard.state[hadron(0)].momentum*etb;
  
    /*    generating momenta (lab frame, with rambo    */
    __details::rambo_generator(rne, hard.state);
    
    //====================================================================
    
    
    /*   calculating the dot prudicts   */
    _M_kin = hard.state;
    
    /*   calculate everything   */
    for(unsigned id = 0; _S_proc_table[id].weight_function != nullptr; ++id)
      (*_S_proc_table[id].weight_function)(_M_states[id], _M_kin);
    
    
    /*   selecting subprocess (NOW BY HAND)  */
    std::size_t ID = 5;
    
    
    /*   the process overall weight   */
    weight *= _M_states[ID].weight;
    
  
    /*    inserting the color states    */
    if(_M_mc_color_sum) _M_states[ID].mc_color_sum(hard.cstate, rng(rne), _S_proc_table[ID].select_color);
    else _M_states[ID].direct_color_sum(hard.cstate, _S_proc_table[ID].select_color);

    
    /*    setting flavors     */
    weight *= (*_S_proc_table[ID].select_flavor)(hard.state, rng(rne));
    
    
    
    //============================ ONLY FOR TESTING ====================================
 
    /*   setting up the color states  */
    for(auto& c : hard.cstate)
    {
      c.make_health_report();
      c.is_color_u3 = false;
      c.color_suppression_index = 0u;
      c.overlap_color_suppression_index = color_suppression_index_u3(c);
    }
  
    /*    check the trace   */
    double su3 = 1.0;
    
    switch(ID)
    {
      case 0:
        su3 = __details::proc5g::su3_tree(_M_kin, 1,2,-1, 0, 3)/6.0;
        break;
      
      case 1:
        su3 = 5.0*__details::proc2q3g::su3_tree(_M_kin, 1,2,-1, 0, 3)/qcd::Na2;
        break;
      case 2:
        su3 = -5.0*__details::proc2q3g::su3_tree(_M_kin, 1,-1,0, 2, 3)/qcd::Na/qcd::Nc/2.0;
        break;
      case 3:
        su3 = -5.0*__details::proc2q3g::su3_tree(_M_kin, 1, 0, -1, 2, 3)/qcd::Na/qcd::Nc/2.0;
        break;
      case 4:
        su3 = -5.0*__details::proc2q3g::su3_tree(_M_kin, -1,1,0, 2, 3)/qcd::Na/qcd::Nc/2.0;
        break;
      case 5:
        su3 = -5.0*__details::proc2q3g::su3_tree(_M_kin, 0, 1, -1, 2, 3)/qcd::Na/qcd::Nc/2.0;
        break;
      case 6:
        su3 = 5.0*__details::proc2q3g::su3_tree(_M_kin, -1, 0, 1, 2, 3)/qcd::Nc2/6.0;
        break;
      case 7:
        su3 = 5.0*__details::proc2q3g::su3_tree(_M_kin, 0, -1, 1, 2, 3)/qcd::Nc2/6.0;
        break;
   }
    
    std::cout<<"|M|^2 = "<<su3<<" ";
    
    weight /= su3;
    //===================================================================================
    
    
    hard.weights.nlo.main = weight;
  }
  

  
  /*  Process 0: g + g --> g + g + g
   */
  static void _S_process_5g(__details::subprocess_state& ck, const __details::dot_products<-1, 3>& kin)
  {
    /*   make room for the matrices   */
    ck.resize(12u);
    
    /*   view these vectors as matrix   */
    span<double, 12u, 12u> prob{ck.cprob}, weight{ck.cweight};
    
    
    /*    calculate all the independent color factors    */
    static constexpr int p1 = -1, p2 = 0, p3 = 1, p4 = 2, p5 = 3;
    
    //-----
    {
      double A1;
      
      prob[0][0]  = A1 = __details::proc5g::A1(kin, p1, p2, p3, p4, p5);
      prob[0][1]  = A1*__details::proc5g::R2(kin, p1, p2, p3, p4, p5);
      prob[0][2]  = A1*__details::proc5g::R3(kin, p1, p2, p3, p4, p5);
      prob[0][3]  = A1*__details::proc5g::R4(kin, p1, p2, p3, p4, p5);
      prob[0][4]  = A1*__details::proc5g::R5(kin, p1, p2, p3, p4, p5);
      prob[0][5]  = A1*__details::proc5g::R6(kin, p1, p2, p3, p4, p5);
      prob[0][6]  = A1*__details::proc5g::R7(kin, p1, p2, p3, p4, p5);
      prob[0][7]  = A1*__details::proc5g::R8(kin, p1, p2, p3, p4, p5);
      prob[0][8]  = A1*__details::proc5g::R9(kin, p1, p2, p3, p4, p5);
      prob[0][9]  = A1*__details::proc5g::R10(kin, p1, p2, p3, p4, p5);
      prob[0][10] = A1*__details::proc5g::R11(kin, p1, p2, p3, p4, p5);
      prob[0][11] = A1*__details::proc5g::R12(kin, p1, p2, p3, p4, p5);
      
      
      prob[1][1]  =  A1 = __details::proc5g::A1(kin, p1, p3, p2, p4, p5);
      prob[1][2]  =  A1*__details::proc5g::R5(kin, p1, p3, p2, p4, p5);
      prob[1][3]  =  A1*__details::proc5g::R6(kin, p1, p3, p2, p4, p5);
      prob[1][4]  =  A1*__details::proc5g::R3(kin, p1, p3, p2, p4, p5);
      prob[1][5]  =  A1*__details::proc5g::R4(kin, p1, p3, p2, p4, p5);
      prob[1][6]  =  A1*__details::proc5g::R12(kin, p1, p3, p2, p4, p5);
      prob[1][7]  = -A1*__details::proc5g::R10(kin, p1, p3, p2, p4, p5);
      prob[1][8]  =  A1*__details::proc5g::R11(kin, p1, p3, p2, p4, p5);
      prob[1][9]  = -A1*__details::proc5g::R8(kin, p1, p3, p2, p4, p5);
      prob[1][10] =  A1*__details::proc5g::R9(kin, p1, p3, p2, p4, p5);
      prob[1][11] =  A1*__details::proc5g::R7(kin, p1, p3, p2, p4, p5);
      
      
      prob[2][2]  =  A1 = __details::proc5g::A1(kin, p1, p3, p4, p2, p5);
      prob[2][3]  =  A1*__details::proc5g::R7(kin, p1, p3, p4, p2, p5);
      prob[2][4]  = -A1*__details::proc5g::R10(kin, p1, p3, p4, p2, p5);
      prob[2][5]  = -A1*__details::proc5g::R9(kin, p1, p3, p4, p2, p5);
      prob[2][6]  = -A1*__details::proc5g::R11(kin, p1, p3, p4, p2, p5);
      prob[2][7]  =  A1*__details::proc5g::R3(kin, p1, p3, p4, p2, p5);
      prob[2][8]  = -A1*__details::proc5g::R12(kin, p1, p3, p4, p2, p5);
      prob[2][9]  = -A1*__details::proc5g::R2(kin, p1, p3, p4, p2, p5);
      prob[2][10] = -A1*__details::proc5g::R4(kin, p1, p3, p4, p2, p5);
      prob[2][11] =  A1*__details::proc5g::R6(kin, p1, p3, p4, p2, p5);
      
      
      prob[3][3]  =  A1 = __details::proc5g::A1(kin, p1, p3, p4, p5, p2);
      prob[3][4]  = -A1*__details::proc5g::R8(kin, p1, p3, p4, p5, p2);
      prob[3][5]  = -A1*__details::proc5g::R5(kin, p1, p3, p4, p5, p2);
      prob[3][6]  = -A1*__details::proc5g::R3(kin, p1, p3, p4, p5, p2);
      prob[3][7]  =  A1*__details::proc5g::R11(kin, p1, p3, p4, p5, p2);
      prob[3][8]  = -A1*__details::proc5g::R2(kin, p1, p3, p4, p5, p2);
      prob[3][9]  = -A1*__details::proc5g::R12(kin, p1, p3, p4, p5, p2);
      prob[3][10] =  A1*__details::proc5g::R6(kin, p1, p3, p4, p5, p2);
      prob[3][11] = -A1*__details::proc5g::R4(kin, p1, p3, p4, p5, p2);
      
      
      prob[4][4]  =  A1 = __details::proc5g::A1(kin, p1, p2, p4, p3, p5);
      prob[4][5]  =  A1*__details::proc5g::R7(kin, p1, p2, p4, p3, p5);
      prob[4][6]  =  A1*__details::proc5g::R6(kin, p1, p2, p4, p3, p5);
      prob[4][7]  =  A1*__details::proc5g::R2(kin, p1, p2, p4, p3, p5);
      prob[4][8]  = -A1*__details::proc5g::R4(kin, p1, p2, p4, p3, p5);
      prob[4][9]  = -A1*__details::proc5g::R3(kin, p1, p2, p4, p3, p5);
      prob[4][10] = -A1*__details::proc5g::R12(kin, p1, p2, p4, p3, p5);
      prob[4][11] = -A1*__details::proc5g::R11(kin, p1, p2, p4, p3, p5);
      
      
      prob[5][5]  =  A1 = __details::proc5g::A1(kin, p1, p2, p4, p5, p3);
      prob[5][6]  = -A1*__details::proc5g::R4(kin, p1, p2, p4, p5, p3);
      prob[5][7]  =  A1*__details::proc5g::R12(kin, p1, p2, p4, p5, p3);
      prob[5][8]  =  A1*__details::proc5g::R6(kin, p1, p2, p4, p5, p3);
      prob[5][9]  = -A1*__details::proc5g::R11(kin, p1, p2, p4, p5, p3);
      prob[5][10] = -A1*__details::proc5g::R2(kin, p1, p2, p4, p5, p3);
      prob[5][11] = -A1*__details::proc5g::R3(kin, p1, p2, p4, p5, p3);
      
      
      prob[6][6]  = A1 = __details::proc5g::A1(kin, p1, p2, p3, p5, p4);
      prob[6][7]  = A1*__details::proc5g::R10(kin, p1, p2, p3, p5, p4);
      prob[6][8]  = A1*__details::proc5g::R5(kin, p1, p2, p3, p5, p4);
      prob[6][9]  = A1*__details::proc5g::R8(kin, p1, p2, p3, p5, p4);
      prob[6][10] = A1*__details::proc5g::R3(kin, p1, p2, p3, p5, p4);
      prob[6][11] = A1*__details::proc5g::R2(kin, p1, p2, p3, p5, p4);
      
      
      prob[7][7]  =  A1 = __details::proc5g::A1(kin, p1, p4, p2, p3, p5);
      prob[7][8]  = -A1*__details::proc5g::R6(kin, p1, p4, p2, p3, p5);
      prob[7][9]  = -A1*__details::proc5g::R5(kin, p1, p4, p2, p3, p5);
      prob[7][10] = -A1*__details::proc5g::R7(kin, p1, p4, p2, p3, p5);
      prob[7][11] = -A1*__details::proc5g::R9(kin, p1, p4, p2, p3, p5);
      
      
      prob[8][8]  =  A1 = __details::proc5g::A1(kin, p1, p2, p5, p3, p4);
      prob[8][9]  =  A1*__details::proc5g::R2(kin, p1, p2, p5, p3, p4);
      prob[8][10] = -A1*__details::proc5g::R10(kin, p1, p2, p5, p3, p4);
      prob[8][11] =  A1*__details::proc5g::R8(kin, p1, p2, p5, p3, p4);
      
      
      prob[9][9]  =  A1 = __details::proc5g::A1(kin, p1, p5, p2, p3, p4);
      prob[9][10] =  A1*__details::proc5g::R8(kin, p1, p5, p2, p3, p4);
      prob[9][11] = -A1*__details::proc5g::R10(kin, p1, p5, p2, p3, p4);
      
      
      prob[10][10] = A1 = __details::proc5g::A1(kin, p1, p3, p5, p2, p4);
      prob[10][11] = A1*__details::proc5g::R5(kin, p1, p3, p5, p2, p4);
      
      
      prob[11][11] = __details::proc5g::A1(kin, p1, p3, p2, p5, p4);
    }
    
    
    /*    calculating the color probabilities and the correspinding weights    */
    ck.weight = ck.complete_probs();
    

    /*   Here we put together everything   */
    static constexpr double cfac = 4.0*meta::pow<5u>(qcd::Cf)/qcd::Na2;
    double sum = 0.0;
    
    for(int i = -1; i < 3; ++i)
      for(int j = i+1; j <= 3; ++j)
        sum += meta::pow<4u>(kin[i][j]);
    
    
    ck.weight *= cfac*sum;
    
    /*   there are three identical gluon in the final state   */
    ck.weight /= 6.0;
  }
  
  
  static void _S_select_color_5g(shower_color_state<hhc>& cstate, std::size_t i, std::size_t j, double w)
  {
    /*    color permutations    */
    static const int perm[24][5] = {
      {-1, 1, 2, 3, 0}, {-1, 2, 1, 3, 0}, {-1, 2, 3, 1, 0}, {-1, 2, 3, 0, 1}, {-1, 1, 3, 2, 0}, {-1, 1, 3, 0, 2},
      {-1, 1, 2, 0, 3}, {-1, 3, 1, 2, 0}, {-1, 1, 0, 2, 3}, {-1, 0, 1, 2, 3}, {-1, 2, 0, 1, 3}, {-1, 2, 1, 0, 3},
      
      /*   reversed    */
      {0, 3, 2, 1, -1}, {0, 3, 1, 2, -1}, {0, 1, 3, 2, -1}, {1, 0, 3, 2, -1}, {0, 2, 3, 1, -1}, {2, 0, 3, 1, -1},
      {3, 0, 2, 1, -1}, {0, 2, 1, 3, -1}, {3, 2, 0, 1, -1}, {3, 2, 1, 0, -1}, {3, 1, 0, 2, -1}, {3, 0, 1, 2, -1}
    };
    
    auto c1 = __details::new_color_state<3>(cstate);
    auto c2 = __details::new_color_state<3>(cstate);
    auto c3 = __details::new_color_state<3>(cstate);
    auto c4 = __details::new_color_state<3>(cstate);
    
    c1->insert_closed_stringL(perm[i], perm[i]+5);
    c1->insert_closed_stringR(perm[j], perm[j]+5);
    c1->weight = w;
    
    c2->insert_closed_stringL(perm[i], perm[i]+5);
    c2->insert_closed_stringR(perm[j+12], perm[j+12]+5);
    c2->weight = -w;
    
    c3->insert_closed_stringL(perm[i+12], perm[i+12]+5);
    c3->insert_closed_stringR(perm[j], perm[j]+5);
    c3->weight = -w;
    
    c4->insert_closed_stringL(perm[i+12], perm[i+12]+5);
    c4->insert_closed_stringR(perm[j+12], perm[j+12]+5);
    c4->weight = w;
  };
    
 
  static double _S_select_flavor_5g(event<hhc>& state, double) {
    state[-1].flavor = state[0].flavor = state[1].flavor = state[2].flavor = state[3].flavor = 0;
    return 1.0;
  }
  
  
  
  /*    Process 1: g + g --> q + qbar + g, where q = d,u,s,c,b
   */
  static void _S_process_2q3g(__details::subprocess_state& ck, int p1, int p2, int p3, int p4, int p5, const __details::dot_products<-1, 3>& kin)
  {
    /*   make room for the matrices   */
    ck.resize(6u);
    
    /*   view these vectors as matrix   */
    span<double, 6u, 6u> prob{ck.cprob}, weight{ck.cweight};
    
    
    /*    calculate all the independent color factors    */
    {
      double A1;
      
      prob[0][0] = A1 = __details::proc2q3g::A1(kin, p1, p2, p3, p4, p5);
      prob[0][1] = A1*__details::proc2q3g::R2(kin, p1, p2, p3, p4, p5);
      prob[0][2] = A1*__details::proc2q3g::R3(kin, p1, p2, p3, p4, p5);
      prob[0][3] = A1*__details::proc2q3g::R4(kin, p1, p2, p3, p4, p5);
      prob[0][4] = A1*__details::proc2q3g::R5(kin, p1, p2, p3, p4, p5);
      prob[0][5] = A1*__details::proc2q3g::R6(kin, p1, p2, p3, p4, p5);

      prob[1][1] = A1 = __details::proc2q3g::A1(kin, p1, p2, p4, p3, p5);
      prob[1][2] = A1*__details::proc2q3g::R5(kin, p1, p2, p4, p3, p5);
      prob[1][3] = A1*__details::proc2q3g::R6(kin, p1, p2, p4, p3, p5);
      prob[1][4] = A1*__details::proc2q3g::R3(kin, p1, p2, p4, p3, p5);
      prob[1][5] = A1*__details::proc2q3g::R4(kin, p1, p2, p4, p3, p5);

      prob[2][2] = A1 = __details::proc2q3g::A1(kin, p1, p3, p2, p4, p5);
      prob[2][3] = A1*__details::proc2q3g::R2(kin, p1, p3, p2, p4, p5);
      prob[2][4] = A1*__details::proc2q3g::R6(kin, p1, p3, p2, p4, p5);
      prob[2][5] = A1*__details::proc2q3g::R5(kin, p1, p3, p2, p4, p5);

      prob[3][3] = A1 = __details::proc2q3g::A1(kin, p1, p3, p4, p2, p5);
      prob[3][4] = A1*__details::proc2q3g::R4(kin, p1, p3, p4, p2, p5);
      prob[3][5] = A1*__details::proc2q3g::R3(kin, p1, p3, p4, p2, p5);

      prob[4][4] = A1 = __details::proc2q3g::A1(kin, p1, p4, p2, p3, p5);
      prob[4][5] = A1*__details::proc2q3g::R2(kin, p1, p4, p2, p3, p5);

      prob[5][5] = __details::proc2q3g::A1(kin, p1, p4, p3, p2, p5);
    }


    /*    calculating the color probabilities and the correspinding weights    */
    ck.weight = ck.complete_probs();
    
    
    /*   Here we put together everything   */
    static constexpr double cfac = qcd::Nc*meta::pow<3>(qcd::Cf);
    
    /* kinematic weight */
    double sum = kin[p1][p2]*kin[p5][p2]*(kin[p1][p2]*kin[p1][p2] + kin[p5][p2]*kin[p5][p2])
    + kin[p1][p3]*kin[p5][p3]*(kin[p1][p3]*kin[p1][p3] + kin[p5][p3]*kin[p5][p3])
    + kin[p1][p4]*kin[p5][p4]*(kin[p1][p4]*kin[p1][p4] + kin[p5][p4]*kin[p5][p4]);
    
    ck.weight *= cfac*sum;
  }

  
  static void _S_process_ggqQg(__details::subprocess_state& ck, const __details::dot_products<-1, 3>& kin)
  {
    _S_process_2q3g(ck, 1, -1, 3, 0, 2, kin);
    
    /*   color averaging for the intial state partons    */
    ck.weight /= qcd::Na2;
  }
  

  static void _S_select_color_ggqQg(shower_color_state<hhc>& cstate, std::size_t i, std::size_t j, double w)
  {
    static const int perm[6][5] = {
      {1, -1, 3, 0, 2}, {1, -1, 0, 3, 2}, {1, 3, -1, 0, 2}, {1, 3, 0, -1, 2}, {1, 0, -1, 3, 2}, {1, 0, 3, -1, 2}
    };
    
    auto c1 = __details::new_color_state<3>(cstate);
    
    c1->insert_open_stringL(perm[i], perm[i]+5);
    c1->insert_open_stringR(perm[j], perm[j]+5);
    c1->weight = w;
  }
  
  
  static double _S_select_flavor_ggqQg(event<hhc>& state, double rho)
  {
    int fq = 1 + static_cast<int>(rho*4.999999);
    
    state[-1].flavor = 0;
    state[ 0].flavor = 0;
    state[ 1].flavor = fq;
    state[ 2].flavor = -fq;
    state[ 3].flavor = 0;
    
    return 5.0;
  }

  
  /*    Process 2: q + g --> q + g + g , where q = d,u,s,c,b
   */
  static void _S_process_qgqgg(__details::subprocess_state& ck, const __details::dot_products<-1, 3>& kin)
  {
    _S_process_2q3g(ck, 1, 2, 3, 0, -1, kin);
    
    /*   color averaging for the intial state partons (- sign: we have an all outgoing calculation)   */
    ck.weight /= -qcd::Na*qcd::Nc;

    /*   there are two identical gluon in the final state   */
    ck.weight /= 2.0;
  }
  
  
  static void _S_select_color_qgqgg(shower_color_state<hhc>& cstate, std::size_t i, std::size_t j, double w)
  {
    static const int perm[6][5] = {
      {1, 2, 3, 0, -1}, {1, 2, 0, 3, -1}, {1, 3, 2, 0, -1}, {1, 3, 0, 2, -1}, {1, 0, 2, 3, -1}, {1, 0, 3, 2, -1}
    };

    auto c1 = __details::new_color_state<3>(cstate);
    
    c1->insert_open_stringL(perm[i], perm[i]+5);
    c1->insert_open_stringR(perm[j], perm[j]+5);
    c1->weight = w;
  }
  
  
  static double _S_select_flavor_qgqgg(event<hhc>& state, double rho)
  {
    int fq = 1 + static_cast<int>(rho*4.999999);
    
    state[-1].flavor = -fq;
    state[ 0].flavor = 0;
    state[ 1].flavor = fq;
    state[ 2].flavor = 0;
    state[ 3].flavor = 0;
    
    return 5.0;
  }


  /*    Process 3: g + q --> q + g + g , where q = d,u,s,c,b
   */
  static void _S_process_gqqgg(__details::subprocess_state& ck, const __details::dot_products<-1, 3>& kin)
  {
    _S_process_2q3g(ck, 1, 2, 3, -1, 0, kin);
    
    /*   color averaging for the intial state partons (- sign: we have an all outgoing calculation)   */
    ck.weight /= -qcd::Na*qcd::Nc;
    
    /*   there are two identical gluon in the final state   */
    ck.weight /= 2.0;
  }
  
  
  static void _S_select_color_gqqgg(shower_color_state<hhc>& cstate, std::size_t i, std::size_t j, double w)
  {
    static const int perm[6][5] = {
      {1, 2, 3, -1, 0}, {1, 2, -1, 3, 0}, {1, 3, 2, -1, 0}, {1, 3, -1, 2, 0}, {1, -1, 2, 3, 0}, {1, -1, 3, 2, 0}
    };
    
    auto c1 = __details::new_color_state<3>(cstate);
    
    c1->insert_open_stringL(perm[i], perm[i]+5);
    c1->insert_open_stringR(perm[j], perm[j]+5);
    c1->weight = w;
  }
  
  
  static double _S_select_flavor_gqqgg(event<hhc>& state, double rho)
  {
    int fq = 1 + static_cast<int>(rho*4.999999);
    
    state[-1].flavor = 0;
    state[ 0].flavor = -fq;
    state[ 1].flavor = fq;
    state[ 2].flavor = 0;
    state[ 3].flavor = 0;
    
    return 5.0;
  }
  
  
  /*    Process 4: qbar + g --> qbar + g + g , where q = d,u,s,c,b
   */
  static void _S_process_QgQgg(__details::subprocess_state& ck, const __details::dot_products<-1, 3>& kin)
  {
    _S_process_2q3g(ck, -1, 2, 3, 0, 1, kin);
    
    /*   color averaging for the intial state partons (- sign: we have an all outgoing calculation)   */
    ck.weight /= -qcd::Na*qcd::Nc;
    
    /*   there are two identical gluon in the final state   */
    ck.weight /= 2.0;
  }
  
  
  static void _S_select_color_QgQgg(shower_color_state<hhc>& cstate, std::size_t i, std::size_t j, double w)
  {
    static const int perm[6][5] = {
      {-1, 2, 3, 0, 1}, {-1, 2, 0, 3, 1}, {-1, 3, 2, 0, 1}, {-1, 3, 0, 2, 1}, {-1, 0, 2, 3, 1}, {-1, 0, 3, 2, 1}
    };
    
    auto c1 = __details::new_color_state<3>(cstate);
    
    c1->insert_open_stringL(perm[i], perm[i]+5);
    c1->insert_open_stringR(perm[j], perm[j]+5);
    c1->weight = w;
  }
  
  
  static double _S_select_flavor_QgQgg(event<hhc>& state, double rho)
  {
    int fq = 1 + static_cast<int>(rho*4.999999);
    
    state[-1].flavor = fq;
    state[ 0].flavor = 0;
    state[ 1].flavor = -fq;
    state[ 2].flavor = 0;
    state[ 3].flavor = 0;
    
    return 5.0;
  }
  

  /*    Process 5: g + qbar --> qbar + g + g , where q = d,u,s,c,b
   */
  static void _S_process_gQQgg(__details::subprocess_state& ck, const __details::dot_products<-1, 3>& kin)
  {
    _S_process_2q3g(ck, 0, 2, 3, -1, 1, kin);
    
    /*   color averaging for the intial state partons (- sign: we have an all outgoing calculation)   */
    ck.weight /= -qcd::Na*qcd::Nc;
    
    /*   there are two identical gluon in the final state   */
    ck.weight /= 2.0;
  }
  
  
  static void _S_select_color_gQQgg(shower_color_state<hhc>& cstate, std::size_t i, std::size_t j, double w)
  {
    static const int perm[6][5] = {
      {0, 2, 3, -1, 1}, {0, 2, -1, 3, 1}, {0, 3, 2, -1, 1}, {0, 3, -1, 2, 1}, {0, -1, 2, 3, 1}, {0, -1, 3, 2, 1}
    };

    auto c1 = __details::new_color_state<3>(cstate);
    
    c1->insert_open_stringL(perm[i], perm[i]+5);
    c1->insert_open_stringR(perm[j], perm[j]+5);
    c1->weight = w;
  }
  
  
  static double _S_select_flavor_gQQgg(event<hhc>& state, double rho)
  {
    int fq = 1 + static_cast<int>(rho*4.999999);
    
    state[-1].flavor = 0;
    state[ 0].flavor = fq;
    state[ 1].flavor = -fq;
    state[ 2].flavor = 0;
    state[ 3].flavor = 0;
    
    return 5.0;
  }
  
  
  
  /*    Process 6: qbar + q --> g + g + g , where q = d,u,s,c,b
   */
  static void _S_process_Qqggg(__details::subprocess_state& ck, const __details::dot_products<-1, 3>& kin)
  {
    _S_process_2q3g(ck, -1, 2, 3, 1, 0, kin);
    
    /*   color averaging for the intial state partons    */
    ck.weight /= qcd::Nc2;
    
    /*   there are three identical gluon in the final state   */
    ck.weight /= 6.0;
  }
  
  
  static void _S_select_color_Qqggg(shower_color_state<hhc>& cstate, std::size_t i, std::size_t j, double w)
  {
    static const int perm[6][5] = {
      {-1, 2, 3, 1, 0}, {-1, 2, 1, 3, 0}, {-1, 3, 2, 1, 0}, {-1, 3, 1, 2, 0}, {-1, 1, 2, 3, 0}, {-1, 1, 3, 2, 0}
    };
    
    auto c1 = __details::new_color_state<3>(cstate);
    
    c1->insert_open_stringL(perm[i], perm[i]+5);
    c1->insert_open_stringR(perm[j], perm[j]+5);
    c1->weight = w;
  }
  
  
  static double _S_select_flavor_Qqggg(event<hhc>& state, double rho)
  {
    int fq = 1 + static_cast<int>(rho*4.999999);
    
    state[-1].flavor = fq;
    state[ 0].flavor = -fq;
    state[ 1].flavor = 0;
    state[ 2].flavor = 0;
    state[ 3].flavor = 0;
    
    return 5.0;
  }
  
  
  /*    Process 7: q + qbar --> g + g + g , where q = d,u,s,c,b
   */
  static void _S_process_qQggg(__details::subprocess_state& ck, const __details::dot_products<-1, 3>& kin)
  {
    _S_process_2q3g(ck, 0, 2, 3, 1, -1, kin);
    
    /*   color averaging for the intial state partons    */
    ck.weight /= qcd::Nc2;
    
    /*   there are three identical gluon in the final state   */
    ck.weight /= 6.0;
  }
  
  
  static void _S_select_color_qQggg(shower_color_state<hhc>& cstate, std::size_t i, std::size_t j, double w)
  {
    static const int perm[6][5] = {
      {0, 2, 3, 1, -1}, {0, 2, 1, 3, -1}, {0, 3, 2, 1, -1}, {0, 3, 1, 2, -1}, {0, 1, 2, 3, -1}, {0, 1, 3, 2, -1}
    };
    
    auto c1 = __details::new_color_state<3>(cstate);
    
    c1->insert_open_stringL(perm[i], perm[i]+5);
    c1->insert_open_stringR(perm[j], perm[j]+5);
    c1->weight = w;
  }
  
  
  static double _S_select_flavor_qQggg(event<hhc>& state, double rho)
  {
    int fq = 1 + static_cast<int>(rho*4.999999);
    
    state[-1].flavor = -fq;
    state[ 0].flavor = fq;
    state[ 1].flavor = 0;
    state[ 2].flavor = 0;
    state[ 3].flavor = 0;
    
    return 5.0;
  }
  
  

  
  
  
  const hhcjets_real::subprocess hhcjets_real::_S_proc_table[] = {

    /*    Process 0: g + g --> g + g + g    */
    {&_S_process_5g, &_S_select_color_5g, &_S_select_flavor_5g},
    
    /*    Process 1: g + g --> q + qbar + g, where q = d,u,s,c,b,   */
    {&_S_process_ggqQg, &_S_select_color_ggqQg, &_S_select_flavor_ggqQg},

    /*    Process 2: q + g --> q + g + g, where q = d,u,s,c,b,   */
    {&_S_process_qgqgg, &_S_select_color_qgqgg, &_S_select_flavor_qgqgg},

    /*    Process 3: g + q --> q + g + g, where q = d,u,s,c,b,   */
    {&_S_process_gqqgg, &_S_select_color_gqqgg, &_S_select_flavor_gqqgg},
    
    /*    Process 4: qbar + g --> qbar + g + g, where q = d,u,s,c,b,   */
    {&_S_process_QgQgg, &_S_select_color_QgQgg, &_S_select_flavor_QgQgg},
    
    /*    Process 5: g + qbar --> qbar + g + g , where q = d,u,s,c,b,   */
    {&_S_process_gQQgg, &_S_select_color_gQQgg, &_S_select_flavor_gQQgg},
    
    /*    Process 6: qbar + q --> g + g + g, where q = d,u,s,c,b,   */
    {&_S_process_Qqggg, &_S_select_color_Qqggg, &_S_select_flavor_Qqggg},
    
    /*    Process 7: q + qbar --> g + g + g, where q = d,u,s,c,b,   */
    {&_S_process_qQggg, &_S_select_color_qQggg, &_S_select_flavor_qQggg},
    

    
    /*    End of list   */
    {nullptr, nullptr, nullptr}
  };
}   //   namespace duct




