/*  Copyright (C) 2012-2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 12/23/12.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __proc_hccjets_real_h__
#define __proc_hccjets_real_h__ 1

//   std includes

//   project headers
#include <bits/duct-model.h>
#include <bits/mc-shower-history.h>
#include <cxx/random.h>




namespace duct {

  namespace __details {
    
    /*    insert a new component into the color state   */
    template<std::size_t Nqcd>
    inline auto new_color_state(shower_color_state<hhc>& c)
    {
      auto cstate = c.push();
      
      cstate->resize(Nqcd);
      cstate->weight = 1.0;
      
      return cstate;
    }
    
    
    /*   Simple matrix class to store dot products
     */
    template<std::ptrdiff_t _Low, std::ptrdiff_t _High>
    class dot_products
    {
      static constexpr std::ptrdiff_t _S_dim = _High-_Low+1;
      static constexpr std::size_t _S_size = _S_dim*_S_dim;
      static constexpr std::ptrdiff_t _S_shift = (_S_dim+1)*_Low;
      
      template<ProcessTag _Proc>
      static constexpr double is_ini(std::ptrdiff_t k)
      {
        if constexpr(is_decay_process<_Proc>)
          return (k == 0 ? -1.0 : 1.0);
        
        return (k == -1 || k == 0 ? -1.0 : 1.0);
      }
      
    public:
      /*   element access   */
      double * operator[](std::ptrdiff_t l) { return _M_s + _S_dim*l - _S_shift;}
      const double * operator[](std::ptrdiff_t l) const { return _M_s + _S_dim*l - _S_shift;}
      
      
      /*   calculate from an event    */
      template<ProcessTag _Proc>
      dot_products& operator=(const event<_Proc>& p)
      {
        for(std::ptrdiff_t l = _Low; l <= _High; ++l)
        {
          /*    diaginal elements    */
          _M_s[_S_dim*l + l - _S_shift] = 0.0;
          
          /*   it is a symmetric matrix    */
          double ll = is_ini<_Proc>(l);
          
          for(std::ptrdiff_t k = l + 1; k <= _High; ++k)
            _M_s[_S_dim*l + k - _S_shift] = _M_s[_S_dim*k + l - _S_shift] = 2.0*ll*is_ini<_Proc>(k)*dot(p[l].momentum, p[k].momentum);
        }
        
        return *this;
      }
      
    private:
      /*   underlying array    */
      double _M_s[_S_size];
    };
    
    
    
    /*   Siple class for representing color probabilities
     */
    struct subprocess_state
    {
      /*    picking random color state according to prob    */
      template<class _Func>
      void mc_color_sum(shower_color_state<hhc>& cstate, const double rho, _Func&& func) const
      {
        std::size_t ic = 0, sz = cprob.size();
        double oprob = cprob[0];
        
        while(ic < sz && oprob < rho)
          oprob += cprob[++ic];
        
        func(cstate, ic/_M_dim, ic%_M_dim, cweight[ic]);
      }
      
      /*   do the direct sum   */
      template<class _Func>
      void direct_color_sum(shower_color_state<hhc>& cstate, _Func&& func) const
      {
        for(std::size_t ic = 0; ic < _M_dim*_M_dim; ic++)
          func(cstate, ic/_M_dim, ic%_M_dim, cweight[ic]*cprob[ic]);
      }
      
      /*   resizing the arrays    */
      void resize(const std::size_t n)
      {
        _M_dim = n;
        cprob.resize(n*n);
        cweight.resize(n*n);
      }
      
      
      /*    calculating the color probabilities and the correspinding weights
       *    the diagonal and upper triangle have to be calculated
       */
      double complete_probs()
      {
        double norm  = 0.0;
        
        for(unsigned l = 0; l < _M_dim; ++l)
        {
          auto ll = _M_dim*l + l;
          
          cweight[ll] = (cweight[ll] < 0.0 ? -1.0 : 1.0);
          cprob[ll] = std::abs(cprob[ll]);
          norm += cprob[ll];
          
          for(unsigned k = l+1; k < _M_dim; ++k)
          {
            auto lk = _M_dim*l + k, kl = _M_dim*k + l;
            
            cweight[lk] = cweight[kl] = (cprob[lk] < 0.0 ? -1.0 : 1.0);
            cprob[lk] = cprob[kl] = std::abs(cprob[lk]);
            norm += 2.0*cprob[lk];
          }
        }
        
        /*   include the normalization in the weight matrix    */
        for(double& x : cprob)
          x /= norm;
        
        return norm;
      }
      
      
      /*   subprocess weight    */
      double weight;
      
      /*   storage for color information    */
      std::vector<double> cprob, cweight;

    private:
      /*   better be foolproof   */
      std::size_t _M_dim = 0;
    };
  
  }  //   namespace __details
  
  
  
  class hhcjets_real
  {
  public:
    explicit hhcjets_real(double Ecm,  bool mccsum = false)
    : _M_hadronA{2212}, _M_hadronB{2212}, _M_s{Ecm*Ecm}, _M_mc_color_sum{mccsum} {}
    
    /*   generate an event    */
    void generate(random_engine_type&, shower_stage<hhc>&);

  private:
    /*   incomming hadrons   */
    int _M_hadronA, _M_hadronB;
    double _M_s;
    
    
  private:
    /*   do color sum by MC (or not)   */
    bool _M_mc_color_sum;
    
    /*    representing the pahse space via invariants   */
    __details::dot_products<-1,3> _M_kin;
    
    
    /*    small class for represnting a subprocess    */
    struct subprocess
    {
      /*    function to calculate the weight of the subprocess    */
      void (*weight_function)(__details::subprocess_state&, const __details::dot_products<-1,3>&);
      
      /*   function to select color state    */
      void (*select_color)(shower_color_state<hhc>&, std::size_t, std::size_t, double);
      
      /*   function to select flavor   */
      double (*select_flavor)(event<hhc>&, double);
    };
    
    /*   organizing subprocesses    */
    static const subprocess _S_proc_table[36];
    __details::subprocess_state _M_states[36];
  };
  
  
}


#endif
