/*  Copyright (C) 2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 9/8/13.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

/*   std headers   */
#include <iostream>
#include <list>
#include <thread>
#include <thread>
#include <chrono>

/*   C headers   */
#include <dlfcn.h>
#include <getopt.h>
#include <sys/types.h>
#include <unistd.h>

/*   project headers    */
#include "main-mod.h"
#include "bits/conf-deductor.h"
#include "bits/duct-module.h"
#include "bits/mc-basic-user.h"
#include "cxx/filesystem.h"
#include "deductor.h"


namespace fs = duct::fs;
using namespace std::chrono_literals;



void blurb()
{
  std::cout
  <<"+-------------------------------------------------------------------------+\n"
  <<"|                       \033[1;31mDeductor release "<<duct::conf::version<<"\033[0m                            |\n"
  <<"|                       Z. Nagy and D.E. Soper                            |\n"
  <<"|                                                                         |\n"
  <<"|             A parton shower Monte Carlo event generator                 |\n"
  <<"|          not including an underlying event or hadronization             |\n"
  <<"|             \033[34mhttp://www.desy.de/~znagy/Site/deductor.html\033[0m                |\n"
  <<"|               \033[34mhttp://pages.uoregon.edu/soper/deductor/\033[0m                  |\n"
  <<"|                                                                         |\n"
  <<"|                                                                         |\n"
  <<"| Deductor is provided without warranty under the terms of the GNU GPLv3. |\n"
  <<"| See COPYING file for details.                                           |\n"
  <<"+-------------------------------------------------------------------------+\n"
  <<std::endl;
}


/*    making a unique filename for the current run    */
static std::string getrunname()
{
  std::string rname = "run";
  
  /*  PID  */
  rname += std::to_string(::getpid());
  
  /*  date & time  */
  char buff[128];
  std::time_t now = std::chrono::system_clock::to_time_t(std::chrono::system_clock::now());
  
  std::strftime(buff, 128, "-%Y-%b-%d-%H%M%S@", std::localtime(&now));
  rname += buff;
  
  /*  host  */
  ::gethostname(buff, 128);
  buff[127] = '\0';
  rname += buff;

  return rname;
}



/*    function to generate the result    */
int main_res(const fs::path &outdir, const fs::path &resdir, void *handle);



static void make_result(const fs::path &outdir, const fs::path &resdir, void *handle)
{
  /*  check if we have already output dir   */
  while(!fs::is_directory(outdir))
    std::this_thread::sleep_for(1.0min);
  
  do
  {
    /*   the body of the function to do the job    */
    main_res(outdir, resdir, handle);
    
    
    std::this_thread::sleep_for(2.0min);
  } while(true);
}






int main_run(int argc, char **argv)
{
  /*   print fancy message   */
  blurb();
  
  /*   module path   */
  std::string module{argv[::optind-1]};
  std::string bundle = module + ".bundle";
  
  /*   parsing the command line   */
  ::option longopt[] = {
    {"parallel",   no_argument, nullptr, 0},
    {"master",   no_argument, nullptr, 0},
    {nullptr, 0, nullptr, 0}
  };
  
  
  /*   argument parsing   */
  bool is_parallel = false, is_master = false;
  
  while(true) {
    int optidx = 0;
    int c = ::getopt_long(argc, argv, "", longopt, &optidx);
    if(c == -1) break;
    
    switch(c) {
      case 0:
        switch(optidx) {
          case 0: is_parallel = true; break;
          case 1: is_master = true; break;
        }
        break;
      case '?': break;
      default: throw; break;
    }
  }
  
  
  
  /*    open the user defined file   */
  fs::path modfile = fs::path{bundle}/(std::string{"/module"}+__SOEXT__);
  void *handle = ::dlopen(modfile.c_str(), RTLD_LAZY);
  
  if(!handle) {
    std::cerr<<"can't open the module "<<modfile<<"!\n";
    throw dlerror();
  }
  
  
  /*   select all the non kernel sub-modules   */
  std::list<duct::module_entry<void(const duct::deductor_input&)> *> calcmods, othermods, inimods, finmods;
  
  /*   checking the global symbol tables    */
  for(auto& m : duct::_G_exported_modules<void(const duct::deductor_input&)>) {
    for(auto sym = m.data(); sym < m.data()+m.size(); ++sym)
      switch(sym->type) {
        case duct::module_type::process: calcmods.push_back(sym);  break;
        case duct::module_type::other:   othermods.push_back(sym); break;
        case duct::module_type::init:    inimods.push_back(sym); break;
        case duct::module_type::finish:  finmods.push_back(sym); break;
        case duct::module_type::none: break; break;
      }
  }
  
  
  /*   generate the name of the run   */
  auto name = getrunname();
  
  /*    print some basic info   */
  std::cout
  <<"\033[1mUser bundle        : \033[0m"<<module<<"\n"
  <<"\033[1mBundle path        : \033[0m"<<bundle<<"\n"
  <<"\033[1mName of the run    : \033[0m"<<name<<"\n";

  

  
  /*   bundle directories   */
  duct::deductor_input input;

  fs::path outdir = fs::path{bundle}/"output";
  fs::path resdir = fs::path{bundle}/"result";
  fs::path mdldir = fs::path{bundle}/"models";

  input.bundle_dir = fs::path{bundle};
  input.run_name = name;
  input.model_dir = mdldir;
  input.output_dir = outdir;
  input.result_dir = resdir;


  /*   the master process keeps genrating the result periodically   */
  std::unique_ptr<std::thread> result_thread;
  
  if(is_master)
    result_thread = std::make_unique<std::thread>(make_result, outdir, resdir, handle);
  
  
  
  /*   runung the init modules first    */
  if(!inimods.empty()) {
    for(auto mod : inimods) {
      std::cout<<"\033[1;31m\nStarting module : \033[0m\"\033[1;32m"<<mod->name<<"\033[0m\"\n";
      mod->func(input);
    }
    
    std::cout<<"\n";
  }

  
  
  /*   runung the other modules first    */
  if(!othermods.empty()) {
    for(auto mod : othermods) {
      std::cout<<"\033[1;31m\nStarting module : \033[0m\"\033[1;32m"<<mod->name<<"\033[0m\"\n";
      mod->func(input);
    }
    
    std::cout<<"\n";
  }
  
  
  
  /*   run the sub-modules parallel   */
  if(is_parallel) {
    std::vector<std::thread> submods;
    
    /*   starting a new thread for every sub-module  */
    for(auto mod : calcmods) {
      std::cout<<"\033[1;31m\nStarting module : \"\033[0m\"\033[1;32m"<< mod->name << "\033[0m\"\n";
    
      input.bundle_dir = fs::path{bundle};
      input.model_dir  = mdldir/mod->name;
      input.output_dir = outdir/mod->name;
      
      fs::create_directories(input.model_dir);
      
      submods.emplace_back([=]() -> void {
        /*   set the global (but it is thread local) inializer of the basic users   */
        duct::basic_user_base::initializer = [=](duct::basic_user_base *u) -> void {
          u->dirname(outdir/mod->name);
          u->runname(name.c_str());
        };
        
        /*   start the calculation   */
        mod->func(input);
      });
    }
    
    /*   the threads join here   */
    for(std::thread& x : submods)
      x.join();
  }
  
  /*   run sub-modules sequentially   */
  else {
    for(auto mod : calcmods)
    {
      input.bundle_dir = fs::path{bundle};
      input.model_dir  = mdldir/mod->name;
      input.output_dir = outdir/mod->name;

      fs::create_directories(input.model_dir);

      /*   set the global (but it is thread local) inializer of the basic users   */
      duct::basic_user_base::initializer = [=](duct::basic_user_base *u) -> void {
        u->dirname(outdir/mod->name);
        u->runname(name.c_str());
      };
      
      /*   start the calculation   */
      std::cout<<"\033[1;31m\nStarting module : \033[0m\"\033[1;32m"<<mod->name<<"\033[0m\"\n";
      mod->func(input);
    }
  }
  
  
  
  /*   runung the final modules last    */
  if(!finmods.empty()) {
    for(auto mod : finmods) {
      std::cout<<"\033[1;31m\nStarting module : \033[0m\"\033[1;32m"<<mod->name<<"\033[0m\"\n";
      mod->func(input);
    }
    
    std::cout<<"\n";
  }
  
  /*   result thread  joins here   */
  if(is_master) result_thread->join();
  
  /*   closing the module   */
  ::dlclose(handle);

  
  
  return 0;
}





//----- print out the help -----
void main_run_help()
{
  std::cout
  <<"Usage: deductor --run module [--parallel] \n\n"
  
  <<"Options of --run mode:\n"
  <<"  module              The name of the module.\n"
  <<"  --parallel          Runs the process modules parallel.\n"
  <<"                      Make sure that the analysers are thread safe!!! \n"
  <<"--------------------------------------------------------------------------------\n"
  <<std::endl;
}


